package com.gvconcepts.test;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.List;

import android.app.Activity;
import android.content.Context;
import android.content.pm.PackageManager;
import android.graphics.ImageFormat;
import android.hardware.Camera;
import android.hardware.Camera.Parameters;
import android.hardware.Camera.PreviewCallback;
import android.media.MediaCodec;
import android.media.MediaCodecInfo;
import android.media.MediaCodecList;
import android.media.MediaFormat;
import android.os.Bundle;
import android.os.Environment;
import android.util.Log;
import android.view.Menu;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.view.View;
import android.widget.Toast;

public class MainActivity extends Activity implements SurfaceHolder.Callback
{
    private static final String TAG = "MainActivity";
    private Camera mCamera;
    private SurfaceView mSurfaceView;
    private SurfaceHolder mHolder;
    private boolean isRecording = false;
    private File mVideoFile;
    private FileOutputStream fos = null;
    private MediaCodec mMediaCodec;
    private ByteBuffer[] inputBuffers;
    private ByteBuffer[] outputBuffers;
    private SurfaceView mSurfaceView2;
    
    @Override
    protected void onCreate(Bundle savedInstanceState)
    {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);
        mSurfaceView = (SurfaceView) findViewById(R.id.surface);
        mSurfaceView2 = (SurfaceView) findViewById(R.id.surface2);
        mHolder = mSurfaceView.getHolder();
        mHolder.addCallback(this);
        
        if (!checkCameraHardware(this))
        {
            toast("Not found camera");
            return;
        }
        Log.i(TAG, "number of cameras-->" + Camera.getNumberOfCameras());
        
        mCamera = getCameraInstance();
        if (mCamera == null)
        {
            Log.e(TAG, "Fail to aquire a camera instance");
            return;
        }
        setCameraParams();
        createFile();
    }
    
    /**
     * 初始化编码器
     */
    private void initCodec()
    {
        try
        {
            fos = new FileOutputStream(mVideoFile, false);
        }
        catch (FileNotFoundException e)
        {
            e.printStackTrace();
        }
        mMediaCodec = MediaCodec.createEncoderByType("video/avc");
        MediaFormat mediaFormat = MediaFormat.createVideoFormat("video/avc",
                320,
                240);
        mediaFormat.setInteger(MediaFormat.KEY_BIT_RATE, 125000);
        mediaFormat.setInteger(MediaFormat.KEY_FRAME_RATE, 15);
        /*
        COLOR_FormatYUV420PackedPlanar    
        int COLOR_FormatYUV420PackedSemiPlanar  
        int COLOR_FormatYUV420Planar    
        int COLOR_FormatYUV420SemiPlanar    
        int COLOR_FormatYUV422PackedPlanar  
        int COLOR_FormatYUV422PackedSemiPlanar  
        int COLOR_FormatYUV422Planar    
        int COLOR_FormatYUV422SemiPlanar    
        int COLOR_FormatYUV444Interleaved   
        int COLOR_QCOM_FormatYUV420SemiPlanar   
        int COLOR_TI_FormatYUV420PackedSemiPlanar*/
        mediaFormat.setInteger(MediaFormat.KEY_COLOR_FORMAT,
                MediaCodecInfo.CodecCapabilities.COLOR_TI_FormatYUV420PackedSemiPlanar);
        mediaFormat.setInteger(MediaFormat.KEY_I_FRAME_INTERVAL, 5);
        mMediaCodec.configure(mediaFormat,
                null,
                null,
                MediaCodec.CONFIGURE_FLAG_ENCODE);
        mMediaCodec.start();
        inputBuffers = mMediaCodec.getInputBuffers();
        outputBuffers = mMediaCodec.getOutputBuffers();
        
        //new ReceiveThread().start();
    }
    
    private void getSupportColorFormat()
    {
        int numCodecs = MediaCodecList.getCodecCount();
        MediaCodecInfo codecInfo = null;
        for (int i = 0; i < numCodecs && codecInfo == null; i++)
        {
            MediaCodecInfo info = MediaCodecList.getCodecInfoAt(i);
            if (!info.isEncoder())
            {
                continue;
            }
            String[] types = info.getSupportedTypes();
            boolean found = false;
            for (int j = 0; j < types.length && !found; j++)
            {
                if (types[j].equals("video/avc"))
                {
                    System.out.println("found");
                    found = true;
                }
            }
            if (!found)
                continue;
            codecInfo = info;
        }
        Log.d(TAG, "Found " + codecInfo.getName() + " supporting "
                + "video/avc");
        
        // Find a color profile that the codec supports
        int colorFormat = 0;
        MediaCodecInfo.CodecCapabilities capabilities = codecInfo.getCapabilitiesForType("video/avc");
        System.out.println("length-" + capabilities.colorFormats.length + "=="
                + Arrays.toString(capabilities.colorFormats));
        for (int i = 0; i < capabilities.colorFormats.length
                && colorFormat == 0; i++)
        {
            int format = capabilities.colorFormats[i];
            switch (format)
            {
                case MediaCodecInfo.CodecCapabilities.COLOR_FormatYUV420Planar:
                    System.out.println("-");
                    break;
                case MediaCodecInfo.CodecCapabilities.COLOR_FormatYUV420PackedPlanar:
                    System.out.println("-");
                    break;
                case MediaCodecInfo.CodecCapabilities.COLOR_FormatYUV420SemiPlanar:
                    System.out.println("-");
                    break;
                case MediaCodecInfo.CodecCapabilities.COLOR_FormatYUV420PackedSemiPlanar:
                    System.out.println("-");
                    break;
                case MediaCodecInfo.CodecCapabilities.COLOR_TI_FormatYUV420PackedSemiPlanar:
                    colorFormat = format;
                    System.out.println("-");
                    break;
                default:
                    Log.d(TAG, "Skipping unsupported color format " + format);
                    break;
            }
        }
        Log.d(TAG, "Using color format " + colorFormat);
    }
    
    /**
     * 对视频数据压缩
     * 
     * @param data
     */
    private void encode(byte[] data)
    {
        int inputBufferIndex = mMediaCodec.dequeueInputBuffer(-1);
        Log.i(TAG, "inputBufferIndex-->" + inputBufferIndex);
        if (inputBufferIndex >= 0)
        {
            ByteBuffer inputBuffer = inputBuffers[inputBufferIndex];
            inputBuffer.clear();
            inputBuffer.put(data);
            mMediaCodec.queueInputBuffer(inputBufferIndex,
                    0,
                    data.length,
                    0,
                    MediaCodec.BUFFER_FLAG_SYNC_FRAME);
        }
        else
        {
            return;
        }
        
        MediaCodec.BufferInfo bufferInfo = new MediaCodec.BufferInfo();
        int outputBufferIndex = mMediaCodec.dequeueOutputBuffer(bufferInfo, 0);
        if (outputBufferIndex >= 0)
        {
            Log.i(TAG, "outputBufferIndex-->" + outputBufferIndex);
            ByteBuffer outBuffer = outputBuffers[outputBufferIndex];
            byte[] outData = new byte[bufferInfo.size];
            outBuffer.get(outData);
            try
            {
                fos.write(outData, 0, outData.length);
                fos.flush();
                Log.i(TAG, "out data -- > " + outData.length);
                mMediaCodec.releaseOutputBuffer(outputBufferIndex, false);
            }
            catch (IOException e)
            {
                e.printStackTrace();
            }
        }
        //mMediaCodec.flush();
    }
    
    private void stopCodec()
    {
        mCamera.setPreviewCallback(null);
        mMediaCodec.stop();
        mMediaCodec.release();
        if (fos != null)
        {
            try
            {
                fos.close();
            }
            catch (IOException e)
            {
                e.printStackTrace();
            }
        }
    }
    
    public void doOnClick(View view)
    {
        switch (view.getId())
        {
            case R.id.start:
                
                //getSupportColorFormat();
                
                initCodec();
                //mCamera.setPreviewCallback(new MyPreviewCallback());
                mCamera.setPreviewCallback(new PreviewCallback()
                {
                    @Override
                    public void onPreviewFrame(byte[] data, Camera camera)
                    {
                        encode(data);
                    }
                });
                break;
            case R.id.stop:
                stopCodec();
                break;
            default:
                break;
        }
    }
    
    private class ReceiveThread extends Thread
    {
        public void run()
        {
            try
            {
                System.out.println("start receive");
                DatagramSocket socket = new DatagramSocket(40000);
                DatagramPacket receivePacket;
                while (true)
                {
                    byte[] data = new byte[1024 * 1024 * 10];
                    receivePacket = new DatagramPacket(data, 0, data.length);
                    socket.receive(receivePacket);
                    System.out.println("receiv data length: "
                            + receivePacket.getLength());
                }
            }
            catch (SocketException e)
            {
                e.printStackTrace();
            }
            catch (IOException e)
            {
                e.printStackTrace();
            }
        }
    };
    
    private class MyPreviewCallback implements PreviewCallback, Runnable
    {
        private byte[] videoData;
        
        public MyPreviewCallback()
        {
            new Thread(this).start();
        }
        
        @Override
        public void run()
        {
            try
            {
                DatagramSocket socket = new DatagramSocket();
                DatagramPacket sendPacket = null;
                while (true)
                {
                    if (videoData != null)
                    {
                        sendPacket = new DatagramPacket(videoData,
                                videoData.length, InetAddress.getLocalHost(),
                                40000);
                        boolean isSended = false;
                        int tryTime = 0;
                        do
                        {
                            try
                            {
                                System.out.println("===start send====");
                                socket.send(sendPacket);
                                isSended = true;
                                System.out.println("data send success");
                            }
                            catch (IOException e)
                            {
                                tryTime++;
                            }
                        } while (!isSended && tryTime < 3);
                        
                        if (!isSended)
                        {
                            Log.e(TAG, "get video data fail");
                            //return;
                        }
                        videoData = null;
                    }
                }
            }
            catch (SocketException e)
            {
                e.printStackTrace();
            }
            catch (UnknownHostException e)
            {
                e.printStackTrace();
            }
        }
        
        @Override
        public void onPreviewFrame(byte[] data, Camera camera)
        {
            if (videoData == null)
                videoData = data;
        }
        
    }
    
    private void createFile()
    {
        mVideoFile = new File(Environment.getExternalStorageDirectory()
                + File.separator + "video.h264");
        if (!mVideoFile.exists())
        {
            try
            {
                mVideoFile.createNewFile();
            }
            catch (IOException e)
            {
                e.printStackTrace();
            }
        }
    }
    
    @Override
    public boolean onCreateOptionsMenu(Menu menu)
    {
        // Inflate the menu; this adds items to the action bar if it is present.
        getMenuInflater().inflate(R.menu.main, menu);
        return true;
    }
    
    private void toast(final String str)
    {
        runOnUiThread(new Runnable()
        {
            @Override
            public void run()
            {
                Toast.makeText(MainActivity.this, str, Toast.LENGTH_SHORT)
                        .show();
            }
        });
    }
    
    /** A safe way to get an instance of the Camera object. */
    private Camera getCameraInstance()
    {
        Camera c = null;
        try
        {
            c = Camera.open();
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        return c;
    }
    
    /**
     * Camera parameter setting
     */
    private void setCameraParams()
    {
        mCamera.setDisplayOrientation(90);
        //mCamera.startFaceDetection();
        
        Parameters parameters = mCamera.getParameters();
        parameters.setPictureFormat(ImageFormat.JPEG);//Sets the image format for picture 设定相片格式为JPEG，默认为NV21 
        parameters.setPreviewFormat(ImageFormat.NV21);//Sets the image format for preview picture，默认为NV21
        parameters.setPictureSize(320, 240);
        parameters.setPreviewSize(320, 240);
        parameters.setFlashMode(Parameters.FLASH_MODE_AUTO);
        parameters.setFocusMode(Parameters.FOCUS_MODE_CONTINUOUS_VIDEO);
        mCamera.setParameters(parameters);
        
        getSupportedCameraParams();
        
    }
    
    //just for dubug test
    private void getSupportedCameraParams()
    {
        Parameters parameters = mCamera.getParameters();
        //        List<String> supportAnbibandList = parameters.getSupportedAntibanding();
        //        List<String> supportColorEffectList = parameters.getSupportedColorEffects();
        //        List<String> supportFlashModeList = parameters.getSupportedFlashModes();
        //        List<String> supportFocusModeList = parameters.getSupportedFocusModes();
        //        List<Size> supportJTSizeList = parameters.getSupportedJpegThumbnailSizes();
        //        List<Integer> supportPFormatList = parameters.getSupportedPictureFormats();
        //        List<Size> supportPSizeList = parameters.getSupportedPictureSizes();
        List<Integer> supportPreFormatsList = parameters.getSupportedPreviewFormats();
        
        System.out.println("support--" + supportPreFormatsList);
        
        //        List<Size> supportPreSizeList = parameters.getSupportedPreviewSizes();
        //        List<String> supportSceneModeList = parameters.getSupportedSceneModes();
        //        List<Size> supportVSizeList = parameters.getSupportedVideoSizes();
        //        List<String> supportWBList = parameters.getSupportedWhiteBalance();
    }
    
    /** Check if this device has a camera */
    private boolean checkCameraHardware(Context context)
    {
        if (context.getPackageManager()
                .hasSystemFeature(PackageManager.FEATURE_CAMERA))
        {
            return true;
        }
        else
        {
            return false;
        }
    }
    
    //SurfaceHolder Callback
    @Override
    public void surfaceCreated(SurfaceHolder holder)
    {
        try
        {
            mCamera.setPreviewDisplay(mHolder);
            mCamera.startPreview();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }
    
    @Override
    public void surfaceChanged(SurfaceHolder holder, int format, int width,
            int height)
    {
        if (mHolder.getSurface() == null)
        {
            return;
        }
        try
        {
            mCamera.stopPreview();
            mCamera.setPreviewDisplay(mHolder);
            mCamera.startPreview();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }
    
    @Override
    public void surfaceDestroyed(SurfaceHolder holder)
    {
        
    }
}
